#!/usr/bin/env python

import astropy.io.fits as pyfits
import numpy

import random, re, itertools


def main():
    split_supervisor(suffix="-err", error_key="flux_aperture15_err")
    split_supervisor(suffix="-psfwideerr", error_key="flux_psf_err_wide")
    split_supervisor(suffix="-cmodelwideerr", error_key="cmodel_flux_err_wide")
    split_supervisor(suffix="-noerr", error_key=None)

def split_supervisor(suffix, error_key):
    error_keys = [error_key, ] if error_key else []
    namemap = [ (error_key, "parent_flux_convolved_2_1_err"), ] if error_key else []

    colnames = [
        "object_id",
        "crossval_weights",
        "redshift",
    ]
    colnames += map(''.join, itertools.product(
        ["g", "r", "i", "z", "y"],
        ["parent_flux_convolved_2_1", ] + error_keys,
    ))

    wholedata = pyfits.open("hsc_s16a_combined_specz_highq_clean_errsim_train_v1.fits")[1].data
    wholedata = numpy.asarray(wholedata)

    for filter in ["g", "r", "i", "z", "y"]:
        # absorption correction & unit transformation (uJy... => erg/s/cm^2/Hz)
        correction = 1e-29*numpy.exp(0.92103403719761845*wholedata["a_"+filter])

        for colname in map(''.join, itertools.product(
            ["parent_flux_convolved_2_1", ] + error_keys,
        )):
            wholedata[filter+colname] *= correction

    # sanitize weight
    # wholedata = wholedata[wholedata["crossval_weights"] > 0]
    wholedata["crossval_weights"] = 15.0*numpy.arcsinh(wholedata["crossval_weights"]/15.0)

    subdata = map_name(wholedata[colnames], namemap)
    crossval_sample = wholedata["crossval_sample"]

    train_set = subdata[crossval_sample < 4]
    validate_set = subdata[crossval_sample == 4]
    test_set = subdata[crossval_sample == 5]

    numpy.savez("split-data{}.npz".format(suffix),
        train_set=train_set,
        validate_set=validate_set,
        test_set=test_set
    )


def map_name(recarray, namemap):
    """
    Change column names according to namemap.
    For each (oldname, newname) in namemap,
    those columns that end with oldname will be renamed to newname.
    """

    if not namemap:
        return recarray

    namemap = dict(namemap)
    regex = re.compile('(?:' + '|'.join(namemap.keys()) + ')$')

    recarray.dtype.names = tuple(
        regex.sub(lambda m: namemap[m.group()], name)
        for name in recarray.dtype.names
    )

    return recarray

if __name__ == "__main__":
    main()
