#!/usr/bin/env python

# Copyright (c) 2016-2017 Sogo Mineo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from lib import pdfguess
from lib import hsc_stdio

import numpy


def startup():
    import argparse
    parser = argparse.ArgumentParser(fromfile_prefix_chars='@',
        description="""
            Predict photo-z.
            Redshifts are estimated from pdf
            such that a loss function is minimized.
        """
    )
    parser.add_argument('pdfmodel',
        help="""
            The model with which to deduce PDF.
        """
    )
    parser.add_argument('input',
        help="""
            Input catalog (fluxes etc).
        """
    )
    parser.add_argument('output_prefix',
        help="""
            Prefix of names of output files.
        """
    )
    parser.add_argument('--pdf-only', action="store_true",
        help="""
            Output probability distributions only.
            Don't output a digest catalog.
        """
    )

    args = parser.parse_args()
    main(**vars(args))


def main(
    pdfmodel,
    input,
    output_prefix,
    pdf_only = False
):
    pdfguesser = pdfguess.PdfGuesser(pdfmodel)
    object_id, xdata = load_input(input)

    zmax  = pdfguesser.zmax
    pdf   = pdfguesser(xdata)

    del pdfguesser

    hsc_stdio.save_pdf    (output_prefix+".fits"    , object_id, pdf, zmax)

    if not pdf_only:
        hsc_stdio.save_catalog(output_prefix+".cat.fits", object_id, pdf, zmax)


def load_input(input):
    catalog = numpy.load(input)

    ignored = set([])
    idfield = "object_id"

    ignored.update([idfield])

    xfields = []

    for key in catalog.dtype.descr:
        if (key[0] not in ignored
        and not key[0].endswith("_err")
        ):
            xfields.append(key)

    dimension = len(xfields)
    n = len(catalog)

    x = numpy.empty(shape=(n, dimension), dtype=float)
    object_id = catalog[idfield].astype(numpy.int64)

    ix = 0
    for key in xfields:
        x[:,ix] = catalog[key[0]]
        ix += 1

    isfinite = numpy.all(numpy.isfinite(x), axis=1)

    x = x[isfinite]
    object_id = object_id[isfinite]

    return object_id, x


if __name__ == "__main__":
    startup()
