#!/usr/bin/env python

# Copyright (c) 2016-2017 Sogo Mineo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy
import tensorflow as tf

import json, os, re, sys

def startup():
    import argparse
    parser = argparse.ArgumentParser(fromfile_prefix_chars='@',
        description="""
            Train a neural net.
        """
    )
    parser.add_argument('layers', metavar='LAYER', nargs='+',
        help="""
            Hidden layers in the neural net.
            Each argument must be an integer expressing the number of neurons
            within the layer, or be "dropout" meaning half of the neurons
            in the previous layer will be randomly killed during training.
        """
    )
    parser.add_argument('--activation', choices=["softplus", "relu"], default="softplus",
        help="""
            Choose an activation function. (default: softplus)
        """
    )
    parser.add_argument('--zmax', type=float, default=7,
        help="""
            Maximum predicted redshift (default: 7)
        """
    )
    parser.add_argument('--nbins', type=int, default=141,
        help="""
            The number of bins in the last layer (default: 141)
        """
    )
    parser.add_argument('--iterations', type=int, default=600000,
        help="""
            Number of learning iterations (default: 600000).
            Learning can be stopped anytime by pressing Ctrl+C
        """
    )
    parser.add_argument('--learnrate-initial', type=float, default=1e-3,
        help="""
            Learning rate's initial value.
        """
    )
    parser.add_argument('--halflife', type=float, default=300000,
        help="""
            Learning rate decreases in accordance with iterations.
            This option specifies the number of iterations at which
            the learning rate gets half.
        """
    )
    parser.add_argument('--batch-size', type=int, default=100,
        help="""
            Batch size in the training loop.
        """
    )
    parser.add_argument('--logging-interval', type=int, default=1000,
        help="""
            Log messages are output during training by this interval.
        """
    )

    args = parser.parse_args()

    try:
        args.layers = sum((text_to_layer(i) for i in args.layers), [])
        if len(args.layers) == 0:
            raise ValueError()
    except StandardError:
        print >> sys.stderr, "Specify hidden layers. e.g., 32 64 128 dropout"
        sys.exit(1)

    main(**vars(args))


def main(
    layers,
    activation = "softplus",
    zmax = 6.0,
    nbins = 128,
    iterations = 600000,
    learnrate_initial = 1e-3,
    halflife = 300000,
    batch_size = 100,
    logging_interval=1000,
):
    path_id = "pdf-{activation}-".format(**locals())\
        + '-'.join(str(i) for i in layers)

    #
    # Load training set
    #
    xdata, dxdata, zdata, wdata = load_data("split-data.npz", "train_set")

    #
    # Input normalizer
    #
    xscale = 1.0 / numpy.median(xdata, axis=0)
    def normalize(x):
        return numpy.arcsinh(x * xscale).astype(numpy.float32)

    # Make one-hot vectors from z
    bin_edges = get_bin_edges(zmax, nbins)
    ydata = to_onehotvector(zdata, bin_edges)

    #
    # Load validating set
    #
    xdata_val, dxdata_val, zdata_val, wdata_val = load_data("split-data.npz", "validate_set")
    ydata_val = to_onehotvector(zdata_val, bin_edges)

    xdata_val = normalize(xdata_val)

    #
    # Create the neural net
    #
    act_funcs = {"softplus": tf.nn.softplus, "relu": tf.nn.relu, };

    sess = tf.InteractiveSession()
    keep_prob = tf.placeholder(tf.float32)

    weights = []
    biases = []

    def new_layer(input, outdim, is_last=False):
        indim = int(input.get_shape()[-1])
        W = tf.Variable(tf.truncated_normal([indim, outdim], stddev=1.0/numpy.sqrt(0.5*indim)))
        b = tf.Variable(tf.zeros([outdim]))

        weights.append(W)
        biases.append(b)

        h = tf.matmul(input, W) + b
        if is_last:
            # h = tf.nn.softmax(h)
            pass
        else:
            h = act_funcs[activation](h)

        return h

    dim = xdata.shape[-1]
    x  = tf.placeholder(tf.float32, shape=[None, dim]) # input
    y_ = tf.placeholder(tf.float32, shape=[None, ydata.shape[-1]])   # supervisor
    w  = tf.placeholder(tf.float32, shape=[None]) # weight
    h = x

    for outdim in layers:
        if outdim == "dropout":
            h = tf.nn.dropout(h, keep_prob)
        else:
            h = new_layer(h, outdim)

    logy = new_layer(h, ydata.shape[-1], is_last=True)
    y = tf.nn.softmax(logy)

    cross_entropy = tf.reduce_sum(tf.nn.softmax_cross_entropy_with_logits(logy, y_) * w)
    log_dz = numpy.log(bin_edges[1] - bin_edges[0])

    #
    # Do training
    #
    learn_rate = tf.placeholder(tf.float32, shape=[]);
    train_step = tf.train.AdamOptimizer(learn_rate).minimize(cross_entropy)
    sess.run(tf.initialize_all_variables())

    with open("loss-"+path_id+".txt", "w", buffering=1) as flog:
        fout = TeeFile(flog, sys.stderr)
        print>>fout, "# pid = {}".format(os.getpid())
        try:
            lambda_ = numpy.log(0.5) / halflife
            for i in xrange(iterations):
                if i % logging_interval == 0:
                    ce = cross_entropy.eval(feed_dict={x: xdata_val, y_: ydata_val, w: wdata_val, keep_prob: 1.0})
                    print>>fout, i, ce + log_dz

                batch_x, batch_y, batch_w = next_batch(batch_size, xdata, dxdata, ydata, wdata)
                # weight must not be overwritten, so that "w = w/sum" is not "w /= sum".
                batch_w = batch_w / numpy.sum(batch_w)

                rate = learnrate_initial * numpy.exp(float(i)*lambda_)
                train_step.run(feed_dict={x: normalize(batch_x), y_: batch_y, w: batch_w, keep_prob: 0.5, learn_rate: rate})

        except KeyboardInterrupt:
            pass

        ce = cross_entropy.eval(feed_dict={x: xdata_val, y_: ydata_val, w: wdata_val, keep_prob: 1.0})
        print>>fout, i+1, ce + log_dz
        print>>fout, "# end marker"
        print>>fout, i+1, numpy.nan

    #
    # Output the validation
    #

    with open("validate-"+path_id+".txt", "w") as fout:
        print >> fout, '# loss = {:.9g} (xdim = {})'.format(ce + log_dz, xdata_val.shape[-1])

    #
    # Output the model
    #
    weights = sess.run(weights)
    biases = sess.run(biases)

    model = {
        "layers"    : layers,
        "activation": activation,
        "xscale"    : xscale.tolist(),
        "zmax"      : float(zmax),
        "nbins"     : int(nbins),
        "weights"   : [W.tolist() for W in weights],
        "biases"    : [b.tolist() for b in biases ],
    }

    json.dump(model, open("model-"+path_id+".json", "w"))


def text_to_layer(text):
    """
    @param text
      * "dropout": dropout layer
      * an integer: softplus(W*h + b) with int(text) output neurons
      * concatenation thereof by '-'. Thus: '32-16-dropout'
    @return
        list of "dropout" or integers
    """

    def to_layer(text):
        if text == "dropout":
            return text

        i = int(text)
        if i > 0:
            return i

        raise ValueError()

    return [to_layer(x) for x in text.split('-')]


def load_data(file_path, key):
    catalog = numpy.load(file_path)[key]

    ignored = set(["object_id"])
    zfield = "redshift"
    wfield = "crossval_weights"

    ignored.add(zfield)
    ignored.add(wfield)

    xfields = []

    for key in catalog.dtype.descr:
        if (key[0] not in ignored
        and not key[0].endswith("_err")
        ):
            xfields.append(key)

    xdim = len(xfields)
    n = len(catalog)

    x  = numpy.empty(shape=(n, xdim), dtype=float)
    dx = numpy.empty(shape=(n, xdim), dtype=float)
    z  = catalog[zfield].astype(float)
    w  = catalog[wfield].astype(float)

    ix = 0
    for key in xfields:
        x[:,ix] = catalog[key[0]]
        if key[0] + "_err" in catalog.dtype.names:
            dx[:,ix] = catalog[key[0]+"_err"]
        else:
            print "# Warning: input catalog does not include " + key[0] + "_err"
            dx[:,ix] = 0.0
        ix += 1

    isfinite = (
        numpy.all(numpy.isfinite(x), axis=1)
      & numpy.all(numpy.isfinite(dx), axis=1)
      & numpy.isfinite(z)
      & numpy.isfinite(w)
      & (w > 0.0)
    )

    x  = x [isfinite]
    dx = dx[isfinite]
    z  = z [isfinite]
    w  = w [isfinite]

    w /= numpy.sum(w)

    if numpy.all(dx == 0.0):
        dx = None

    return x, dx, z, w


def get_bin_edges(zmax, nbins):
    """
    Let z[n] the representative point of the n-th bin:
        z[0] = 0, z[nbins-1] = zmax.

    The returned "bin edges" are defined such that
        0.5*(edges[:-1] + edges[1:]) == z
    """

    i = numpy.arange(nbins+1).astype(float)
    edges = zmax * (i - 0.5) / (nbins-1)

    return edges


def to_onehotvector(value, bin_edges):
    value = numpy.asarray(value, dtype=float)
    if value.shape:
        is_vector = True
    else:
        is_vector = False
        value = value.reshape((1,))

    digit = numpy.digitize(value, bin_edges[1:-1])

    onehotvec = numpy.zeros(shape=(len(value), len(bin_edges)-1), dtype=numpy.float32)
    onehotvec[numpy.arange(len(value)), digit] = 1.0

    if is_vector:
        return onehotvec
    else:
        return onehotvec[0]


def next_batch(n, x, dx, *others):
    indices = numpy.random.choice(len(x), size=n, replace=False)

    x  = x[indices]
    others = tuple(arr[indices] for arr in others)

    if dx is not None:
        dx = dx[indices]
        x = numpy.copy(x)
        x += numpy.random.normal(size=dx.shape) * dx

    return (x,) + others


class TeeFile(object):
    def __init__(self, file, *files):
        self.files = [file]
        self.files.extend(files)

    def __getattr__(self, name):
        attr = [getattr(f, name) for f in self.files]
        if hasattr(attr[0], "__call__"):
            def wrapper(*x, **y):
                ret = attr[0](*x, **y)
                for a in attr[1:]:
                    a(*x, **y)
                return ret
            return wrapper
        else:
            return attr[0]


if __name__ == "__main__":
    startup()
