#!/bin/echo This is not executable but a library:

# Copyright (c) 2016-2017 Sogo Mineo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import pdfdigest
from .pyfits_proxy import pyfits

import numpy

import sys


def load_pdf(filename):
    """
    @return (object_id, pdf, zmax)
        zmax is the representative value of
        the right-most bin. It is not the
        right edge of the bin.
    """
    fits = pyfits.open(filename, uint=True)

    data = fits["DATA"].data
    bins = fits["BINS"].data

    object_id = data["ID"]
    pdf = data["PDF"]
    z   = bins["BINS"]

    return object_id, pdf, z[-1]


def save_pdf(filename, object_id, pdf, zmax):
    n, nbins = pdf.shape
    bin_edges = pdfdigest.get_bin_edges(zmax, nbins)

    z = 0.5*(bin_edges[:-1]+bin_edges[1:])

    hdus = [pyfits.PrimaryHDU()]

    hdus.append(pyfits.BinTableHDU.from_columns(
        [   array_to_column("ID", object_id),
            array_to_column("PDF", pdf),
        ],
        header=pyfits.Header({
            "EXTNAME" : "DATA",
            "z_min"   : 0.0,
            "z_max"   : float(zmax),
            "delta_z" : float(zmax) / (nbins - 1),
        })
    ))

    hdus.append(pyfits.BinTableHDU.from_columns(
        [   array_to_column("BINS", z.astype(numpy.float32)),
        ],
        header=pyfits.Header({
            "EXTNAME" : "BINS",
        })
    ))

    pyfits.HDUList(hdus).writeto(filename, clobber=True)


def save_catalog(filename, object_id, pdf, zmax):
    n, nbins = pdf.shape
    digestor = pdfdigest.PdfDigestor(pdf, zmax)

    errbound = digestor.get_error_boundaries()
    zm2 = errbound[:,0]
    zm1 = errbound[:,1]
    zp0 = errbound[:,2]
    zp1 = errbound[:,3]
    zp2 = errbound[:,4]

    dtype = [
        ("ID"                 , numpy.int64  ),
        ("PHOTOZ_MEAN"        , numpy.float32),
        ("PHOTOZ_MODE"        , numpy.float32),
        ("PHOTOZ_MEDIAN"      , numpy.float32),
        ("PHOTOZ_BEST"             , numpy.float32),
        ("PHOTOZ_MC"          , numpy.float32),
        ("PHOTOZ_CONF_MEAN"   , numpy.float32),
        ("PHOTOZ_CONF_MODE"   , numpy.float32),
        ("PHOTOZ_CONF_MEDIAN" , numpy.float32),
        ("PHOTOZ_CONF_BEST"        , numpy.float32),
        ("PHOTOZ_RISK_MEAN"   , numpy.float32),
        ("PHOTOZ_RISK_MODE"   , numpy.float32),
        ("PHOTOZ_RISK_MEDIAN" , numpy.float32),
        ("PHOTOZ_RISK_BEST"        , numpy.float32),
        ("PHOTOZ_STD_MEAN"    , numpy.float32),
        ("PHOTOZ_STD_MODE"    , numpy.float32),
        ("PHOTOZ_STD_MEDIAN"  , numpy.float32),
        ("PHOTOZ_STD_BEST"         , numpy.float32),
        ("PHOTOZ_ERR68_MIN"   , numpy.float32),
        ("PHOTOZ_ERR68_MAX"   , numpy.float32),
        ("PHOTOZ_ERR95_MIN"   , numpy.float32),
        ("PHOTOZ_ERR95_MAX"   , numpy.float32),
    ]

    table = numpy.empty(n, dtype)

    table["ID"] = object_id

    table["PHOTOZ_MEAN"]   = zmean = digestor.get_mean()
    table["PHOTOZ_MODE"]   = zmode = digestor.get_mode()
    table["PHOTOZ_MEDIAN"] = zp0
    table["PHOTOZ_MC"]     = digestor.get_montecarlo()

    table["PHOTOZ_CONF_MEAN"]   = digestor.get_confidence(zmean)
    table["PHOTOZ_CONF_MODE"]   = digestor.get_confidence(zmode)
    table["PHOTOZ_CONF_MEDIAN"] = digestor.get_confidence(zp0)

    table["PHOTOZ_STD_MEAN"]   = digestor.get_stddev(zmean)
    table["PHOTOZ_STD_MODE"]   = digestor.get_stddev(zmode)
    table["PHOTOZ_STD_MEDIAN"] = digestor.get_stddev(zp0)

    table["PHOTOZ_ERR68_MIN"] = zm1
    table["PHOTOZ_ERR68_MAX"] = zp1
    table["PHOTOZ_ERR95_MIN"] = zm2
    table["PHOTOZ_ERR95_MAX"] = zp2

    zbest, risk = digestor.get_point_estimate()
    table["PHOTOZ_BEST"]      = zbest
    table["PHOTOZ_RISK_BEST"] = risk
    table["PHOTOZ_CONF_BEST"] = digestor.get_confidence(zbest)
    table["PHOTOZ_STD_BEST"]   = digestor.get_stddev(zbest)

    table["PHOTOZ_RISK_MEAN"]   = digestor.get_risk(zmean)
    table["PHOTOZ_RISK_MODE"]   = digestor.get_risk(zmode)
    table["PHOTOZ_RISK_MEDIAN"] = digestor.get_risk(zp0)

    pyfits.BinTableHDU(table, uint=True).writeto(filename, clobber=True)

def array_to_column(name, arr):
    """
    @param name: str
    @param arr: numpy.array
    @return: pyfits.Column

    Such a function as this one must be in pyfits,
    but I cannot find one.
    """

    type_dict = {
        numpy.bool      : 'L',
        numpy.uint8     : 'B',
        numpy.uint16    : 'I',
        numpy.uint32    : 'J',
        numpy.uint64    : 'K',
        numpy.int8      : 'B',
        numpy.int16     : 'I',
        numpy.int32     : 'J',
        numpy.int64     : 'K',
        numpy.float16   : 'E',
        numpy.float32   : 'E',
        numpy.float64   : 'D',
        numpy.complex64 : 'C',
        numpy.complex128: 'M',
    }

    tzero_dict = {
        numpy.int8      : -128,
        numpy.uint16    : 32768,
        numpy.uint32    : 2147483648,
        numpy.uint64    : 9223372036854775808,
    }

    form  = type_dict[arr.dtype.type]
    tzero = tzero_dict.get(arr.dtype.type)

    shape = arr.shape[1:]
    nelems = 1
    for d in shape:
        nelems *= d

    if nelems != 1:
        form = "{nelems:d}{form}".format(**locals())

    if len(shape) > 1:
        tdim = ('({}' + ',{}'*(len(shape)-1) + ')').format(*reversed(shape))
    else:
        tdim = None

    return pyfits.Column(name, form, bzero=tzero, dim=tdim, array=arr)
