#!/bin/echo This is not executable but a library:

# Copyright (c) 2016-2017 Sogo Mineo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy
import itertools, json, math


class PdfDigestor(object):
    def __init__(self, pdf, zmax):
        """
        @param pdf: numpy 2d-array
            pdf[i][k] = p_i(z_k) dz
        @param zmax: float
            The representative value of the right-most bin.
        """
        pdf = numpy.array(pdf, dtype=float)
        n, nbins = pdf.shape

        pdf /= numpy.sum(pdf, axis=1).reshape(n, 1)

        self.pdf = pdf
        self.bin_edges = get_bin_edges(zmax, pdf.shape[1])
        self.cum_p = None # computed on demand
        self.risk  = None # computed on demand

    def get_point_estimate(self):
        """
        z* that minimizes \int dz P(z) loss(z*, z)
        """
        risk, z_guess = self.get_risk_distribution()
        nguesses = len(z_guess)

        peak = numpy.argmin(risk, axis=1)
        minrisk = risk[numpy.arange(len(peak)), peak]

        is_peak_l = (peak == 0)
        is_peak_r = (peak == nguesses - 1)
        is_peak_m = ~(is_peak_l | is_peak_r)

        z_out = numpy.empty(risk.shape[0])

        # left side
        z_out[is_peak_l] = z_guess[0]

        # right side
        z_out[is_peak_r] = z_guess[-1]

        # middle
        peak_sel = peak[is_peak_m]
        p0 = risk[is_peak_m, peak_sel-1]
        p1 = risk[is_peak_m, peak_sel]
        p2 = risk[is_peak_m, peak_sel+1]
        z0 = z_guess[peak_sel-1]
        z1 = z_guess[peak_sel]
        z2 = z_guess[peak_sel+1]

        # interpolate (z,p) with 2nd-order polynomial, and calc its peak position
        z_out[is_peak_m] = (
            ((z1**2 - z2**2)*p0 + (z2**2 - z0**2)*p1 + (z0**2 - z1**2)*p2)
                /
            (2*((z1 - z2)*p0 + (z2 - z0)*p1 + (z0 - z1)*p2))
        )

        return z_out, minrisk

    def get_risk(self, z):
        risk, z_guess = self.get_risk_distribution()
        n, nbins = risk.shape

        risk_at_z = numpy.empty(n, dtype=float)

        for i in xrange(n):
            risk_at_z[i] = numpy.interp(z[i], z_guess, risk[i])

        return risk_at_z

    def get_mode(self):
        pdf = self.pdf
        bin_edges = self.bin_edges

        z_true = 0.5*(bin_edges[:-1] + bin_edges[1:])
        nbins = len(z_true)

        peak = numpy.argmax(pdf, axis=1)

        is_peak_l = (peak == 0)
        is_peak_r = (peak == nbins - 1)
        is_peak_m = ~(is_peak_l | is_peak_r)

        z_out = numpy.empty(pdf.shape[0])

        # left side
        z_out[is_peak_l] = z_true[0]

        # right side
        z_out[is_peak_r] = z_true[-1]

        # middle
        peak_sel = peak[is_peak_m]
        p0 = pdf[is_peak_m, peak_sel-1]
        p1 = pdf[is_peak_m, peak_sel]
        p2 = pdf[is_peak_m, peak_sel+1]
        z0 = z_true[peak_sel-1]
        z1 = z_true[peak_sel]
        z2 = z_true[peak_sel+1]

        # interpolate (z,p) with 2nd-order polynomial, and calc its peak position
        z_out[is_peak_m] = (
            ((z1**2 - z2**2)*p0 + (z2**2 - z0**2)*p1 + (z0**2 - z1**2)*p2)
                /
            (2*((z1 - z2)*p0 + (z2 - z0)*p1 + (z0 - z1)*p2))
        )

        return z_out

    def get_mean(self):
        pdf = self.pdf
        bin_edges = self.bin_edges

        z = 0.5*(bin_edges[:-1] + bin_edges[1:])
        z = z.reshape(1, len(z))

        return numpy.sum(z*pdf, axis=1)

    def get_stddev(self, zmean):
        pdf = self.pdf
        bin_edges = self.bin_edges

        z = 0.5*(bin_edges[:-1] + bin_edges[1:])
        z = z.reshape(1, len(z))

        zmean = zmean.reshape(len(zmean), 1)
        dev = numpy.square(z - zmean)

        return numpy.sqrt(numpy.sum(dev*pdf, axis=1))

    def get_error_boundaries(self):
        """
        Get error boundaries (-2sigma, -sigma, median, +sigma, +2sigma)
        from cumulative probability.
        """
        cum_p = self.get_cumulative_probability()
        bin_edges = self.bin_edges

        if False:
            sigmas = [-2, -1, 0, 1, 2]
            cum_p_at_sigmas = numpy.array([0.5*math.erfc(-s/math.sqrt(2)) for s in sigmas], dtype=float)
        else:
            cum_p_at_sigmas = numpy.array([0.025, 0.16, 0.5, 0.84, 0.975], dtype=float)

        errbound = numpy.empty(shape=(cum_p.shape[0], len(cum_p_at_sigmas)))
        for i in xrange(cum_p.shape[0]):
            errbound[i] = numpy.interp(cum_p_at_sigmas, cum_p[i], bin_edges)

        return errbound

    def get_confidence(self, zguess):
        cum_p = self.get_cumulative_probability()
        bin_edges = self.bin_edges

        width = 0.03 * (1.0 + zguess)

        n = cum_p.shape[0]
        lower_upper = numpy.empty((n, 2), dtype=float)
        lower_upper[:, 0] = zguess - width
        lower_upper[:, 1] = zguess + width

        for i in xrange(n):
            # [cum_p(lb), cum_p(ub)] <= [lb, up]
            lower_upper[i] = numpy.interp(lower_upper[i], bin_edges, cum_p[i])

        return lower_upper[:, 1] - lower_upper[:, 0]

    def get_montecarlo(self):
        cum_p = self.get_cumulative_probability()
        bin_edges = self.bin_edges

        n = cum_p.shape[0]
        p = numpy.random.random(size=n)

        zmc = numpy.empty(shape=(n,))
        for i in xrange(n):
            zmc[i] = numpy.interp(p[i], cum_p[i], bin_edges)

        return zmc

    def get_cumulative_probability(self):
        if self.cum_p is None:
            self.cum_p = pdf_to_cumulative_probability(self.pdf)
        return self.cum_p

    def get_risk_distribution(self):
        if self.risk is None:
            risk, z_guess = pdf_to_risk(self.pdf, self.bin_edges)
            self.risk    = risk
            self.z_guess = z_guess

        return self.risk, self.z_guess


def get_bin_edges(zmax, nbins):
    """
    Let z[n] the representative point of the n-th bin:
        z[0] = 0, z[nbins-1] = zmax.

    The returned "bin edges" are defined such that
        0.5*(edges[:-1] + edges[1:]) == z
    """

    i = numpy.arange(nbins+1).astype(float)
    edges = zmax * (i - 0.5) / (nbins-1)

    return edges


def pdf_to_cumulative_probability(pdf):
    """
    Compute comulative probability
        cum_p[i][k] = \int_{bin_edges[0]}^{bin_edges[k]} dz P_i(z).

    @param pdf: numpy 2D array (n, nbins)
        pdf[i][k] = \int_{bin_edges[k]}^{bin_edges[k+1]} dz P_i(z)

    @return cum_p
    """
    cum_p = numpy.empty(shape=(pdf.shape[0], pdf.shape[1]+1), dtype=float)
    cum_p[:,0]  = 0
    numpy.cumsum(pdf, axis=1, out=cum_p[:,1:])
    cum_p /= cum_p[:,-1].reshape((pdf.shape[0], 1))
    return cum_p


def pdf_to_risk(pdf, bin_edges):
    z_true = 0.5*(bin_edges[:-1] + bin_edges[1:])
    zmin, zmax = z_true[0], z_true[-1]

    nguesses = int((zmax - zmin) / 0.01 + 0.5) + 1
    z_guess = zmin + numpy.arange(nguesses).astype(float) / (nguesses-1) * (zmax - zmin)

    z_true  = z_true.reshape(len(z_true),1)
    z_guess = z_guess.reshape(1,nguesses)

    loss_matrix = numpy.asmatrix(loss(z_guess, z_true))
    risk = numpy.asarray(numpy.asmatrix(pdf) * loss_matrix)

    return risk, z_guess.reshape(nguesses)


def mean_loss(z, z_true, weight=None):
    return numpy.average(loss(z, z_true), weights=weight)


def loss(z, z_true):
    def cauchy(x, width):
        inv_ww = 1.0/(width*width)
        return 1.0 - 1.0 / (1.0 + inv_ww*numpy.square(x))

    return cauchy((z_true - z) / (1.0 + z_true), 0.15)
