#!/usr/bin/env python

# Copyright (c) 2016-2017 Sogo Mineo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import glob
import collections
import cStringIO
import re
import sys
import subprocess
import time

import numpy


def startup():
    main(**vars(argparse().parse_args()))


def argparse():
    import argparse
    parser = argparse.ArgumentParser(
        description="Plot losses as functions of #iteration."
    )
    parser.add_argument('loss_log', metavar='loss-log', nargs="*",
        help="""
            File(s) named 'loss-*.txt'.
            If none specified, the current directory is searched.
        """)
    parser.add_argument('--max-finished', metavar='N', type=int, default=1,
        help="""
            Maximum number of finished learnings to show. (-1 for no limit)
        """)
    parser.add_argument('--max-unfinished', metavar='N', type=int, default=-1,
        help="""
            Maximum number of unfinished learnings to show. (-1 for no limit)
        """)

    return parser


def main(loss_log, max_finished=1, max_unfinished=-1):
    gnuplot = Gnuplot()
    pipe = gnuplot.get_pipe()

    while True:
        plot(pipe, loss_log, max_finished, max_unfinished)
        time.sleep(10)


class Gnuplot(object):
    def __init__(self):
        self.pipe = subprocess.Popen(["gnuplot"], stdin=subprocess.PIPE)

    def get_pipe(self):
        return self.pipe.stdin


LogInfo = collections.namedtuple("LogInfo", ["path", "is_finished", "min_loss", "last_loss", "range_loss"])

def read_log(loss_log):
    text = open(loss_log).read()
    # remove incomplete line
    text = re.sub(r"[^\n]*\Z", r"", text)

    table = numpy.loadtxt(cStringIO.StringIO(text), ndmin=1, dtype=[("iter",int),("loss",float)])
    if len(table) == 0:
        return LogInfo(loss_log, False, numpy.inf, numpy.inf)

    is_finished = numpy.isnan(table[-1]["loss"])
    if is_finished:
        table = table[:-1]

    min_loss    = numpy.min(table["loss"])
    last_loss   = table[-1]["loss"]
    q1, q3 = numpy.percentile(table["loss"], [25.0, 75.0])
    range_loss  = 2.0 * (q3 - q1)

    return LogInfo(loss_log, is_finished, min_loss, last_loss, range_loss)


def plot(fout, loss_logs, max_finished=1, max_unfinished=-1):
    if not loss_logs:
        loss_logs = glob.iglob("loss-pdf-softplus-*")

    logs = [read_log(path) for path in loss_logs]

    finished   = [log for log in logs if log.is_finished]
    unfinished = [log for log in logs if not log.is_finished]

    if max_finished > 0:
        finished.sort(key=lambda x: x.last_loss)
        finished = finished[0:max_finished]
    elif max_finished == 0:
        finished = []

    if max_unfinished > 0:
        unfinished.sort(key=lambda x: x.last_loss)
        unfinished = unfinished[0:max_unfinished]
    elif max_unfinished == 0:
        unfinished = []

    logs = finished + unfinished

    logs.sort(key=lambda x: -x.last_loss)
    loss_logs = [log.path for log in logs]

    miny = min(log.min_loss for log in logs)
    maxy = miny + logs[-1].range_loss

    miny = numpy.floor(miny*100.0) / 100.0
    maxy = numpy.ceil(maxy*100.0) / 100.0

    print>>fout, "set yrange [{}:{}]".format(miny, maxy)

    prefix = "plot"
    for path in loss_logs:
        print>>fout, '{prefix} "{path}" with lines title "{path}"'.format(**locals()),
        prefix = ","

    print>>fout, ""
    fout.flush()


if __name__ == "__main__":
    startup()
