#!/usr/bin/env python

# Copyright (c) 2016-2017 Sogo Mineo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from lib import pdfguess
from lib import pdfdigest

import numpy

import itertools, json, math, re, sys, os

def startup():
    import argparse
    parser = argparse.ArgumentParser(fromfile_prefix_chars='@',
        description="""
            Test a model.
            Redshifts are estimated from pdf
            such that a loss function is minimized.
        """
    )
    parser.add_argument('pdfmodel',
        help="""
            The model with which to deduce PDF.
        """
    )
    parser.add_argument('--dataset', choices=["train", "validate", "test"], default="validate",
        help="""
            Which dataset to use (default: validate)
        """
    )
    parser.add_argument('--enable-error', default=False, action="store_true",
        help="""
            Enable perturbation with errors (default: disabled)
        """
    )
    parser.add_argument('--disable-error', default=False, action="store_false",
        dest="enable_error",
        help="""
            Disable perturbation with errors (default: disabled)
        """
    )

    args = parser.parse_args()

    main(**vars(args))


def main(
    pdfmodel,
    dataset,
    enable_error,
):
    if dataset.startswith("train"):
        dataset_key = "train_set"
        dataset = "train"
    elif dataset.startswith("valid"):
        dataset_key = "validate_set"
        dataset = "validate"
    elif dataset.startswith("test"):
        dataset_key = "test_set"
        dataset = "test"
    else:
        raise ArgumentError("dataset must be 'train' or 'validate' or 'test'")

    object_id, xdata, dxdata, zdata, wdata = load_data("split-data.npz", dataset_key)
    if enable_error:
        xdata += dxdata*numpy.random.normal(size=dxdata.shape)

    pdfguesser = pdfguess.PdfGuesser(pdfmodel)

    zmax  = pdfguesser.zmax
    nbins = pdfguesser.nbins
    pdf   = pdfguesser(xdata)

    del pdfguesser

    path_id = dataset \
        + '-' + os.path.splitext(os.path.basename(pdfmodel))[0]

    digestor = pdfdigest.PdfDigestor(pdf, zmax)

    zbest, risk = digestor.get_point_estimate()
    zmode = digestor.get_mode()

    zmean   = digestor.get_mean()
    zstddev = digestor.get_stddev(zmean)

    errbound = digestor.get_error_boundaries()
    zm2 = errbound[:,0]
    zm1 = errbound[:,1]
    zp0 = errbound[:,2]
    zp1 = errbound[:,3]
    zp2 = errbound[:,4]

    zmc = digestor.get_montecarlo()

    zbest_conf = digestor.get_confidence(zbest)
    zmode_conf = digestor.get_confidence(zmode)
    zmean_conf = digestor.get_confidence(zmean)
    zp0_conf   = digestor.get_confidence(zp0)

    #
    # Output
    #

    with open("catalog-"+path_id+".txt", "w") as fout:
        print >> fout, '# loss = {:.9g}'.format(pdfdigest.mean_loss(zbest, zdata, wdata))

        columns = [
            # name        header    cell      value
            ["object_id" , "%18s", "%20d"  , object_id ],
            ["z_true"    , "%15s", "%15.8e", zdata     ],
            ["z_guess"   , "%15s", "%15.8e", zbest     ],
            ["risk"      , "%15s", "%15.8e", risk      ],
            ["z_mode"    , "%15s", "%15.8e", zmode     ],
            ["z_mc"      , "%15s", "%15.8e", zmc       ],
            ["z_2.5%"    , "%15s", "%15.8e", zm2       ],
            ["z_16%"     , "%15s", "%15.8e", zm1       ],
            ["z_50%"     , "%15s", "%15.8e", zp0       ],
            ["z_84%"     , "%15s", "%15.8e", zp1       ],
            ["z_97.5%"   , "%15s", "%15.8e", zp2       ],
            ["z_mean"    , "%15s", "%15.8e", zmean     ],
            ["stddev"    , "%15s", "%15.8e", zstddev   ],
            ["conf_guess", "%15s", "%15.8e", zbest_conf],
            ["conf_mode" , "%15s", "%15.8e", zmode_conf],
            ["conf_mean" , "%15s", "%15.8e", zmean_conf],
            ["conf_50%"  , "%15s", "%15.8e", zp0_conf  ],
        ]

        print >> fout, "# " + " ".join(col[1] for col in columns) % tuple(
            col[0] for col in columns
        )

        rowfmt = " ".join(col[2] for col in columns)
        for row in itertools.izip(*(col[3] for col in columns)):
            print >> fout, rowfmt % row

    plot_hist2d(
        "hist-"+path_id+".pdf",
        wdata, zdata, zbest, xlim=[0,6], ylim=[0,6]
    )

    plot_outlierrate("outliers-"+path_id+".pdf",
        wdata, zdata, zbest, risk, zbest_conf
    )

    pdffile = numpy.empty(len(object_id), dtype=[("object_id", numpy.int64), ("pdf", numpy.float16, pdf.shape[1:])])
    pdffile["object_id"] = object_id
    pdffile["pdf"] = pdf

    numpy.save("pdf-"+path_id+".npy", pdffile)


def plot_hist2d(file_path, weight, x, y, xlim=None, ylim=None):
    from matplotlib import pyplot
    import matplotlib.colors

    pyplot.rcParams.update({
        'axes.labelsize': 10,
        'font.size': 10,
        'legend.fontsize': 10,
        'xtick.labelsize': 10,
        'ytick.labelsize': 10,
        'font.family': "serif",
    })

    if xlim:
        selection = (xlim[0] <= x) & (x <= xlim[1])
        x, y, weight = x[selection], y[selection], weight[selection]
    if ylim:
        selection = (ylim[0] <= y) & (y <= ylim[1])
        x, y, weight = x[selection], y[selection], weight[selection]

    if xlim:
        xq1, xq2, xq3 = numpy.percentile(x, [25, 50, 75])
    else:
        xq0, xq1, xq2, xq3, xq4 = numpy.percentile(x, [0, 25, 50, 75, 100])
        xlim = [xq0, xq4]

    if ylim:
        yq1, yq2, yq3 = numpy.percentile(y, [25, 50, 75])
    else:
        yq0, yq1, yq2, yq3, yq4 = numpy.percentile(y, [0, 25, 50, 75, 100])
        ylim = [yq0, yq4]

    weight *= len(weight) / numpy.sum(weight)

    xsigma = 0.74 * (xq3 - xq1)
    ysigma = 0.74 * (yq3 - yq1)

    n = len(x)
    xh = 3.5 * xsigma / (n+1.0)**0.25
    yh = 3.5 * ysigma / (n+1.0)**0.25

    xbins = int(numpy.ceil((xlim[1] - xlim[0]) / xh))
    ybins = int(numpy.ceil((ylim[1] - ylim[0]) / yh))

#    fig = pyplot.figure(figsize=(5.5, 4.5))
    fig = pyplot.figure(figsize=(4.4, 3.6)) # smaller size for papers
    ax = fig.add_subplot(1,1,1)
    ax.axis("square")

    counts, xedges, yedges, im = ax.hist2d(
        x, y, weights=weight, bins=[xbins, ybins], range=[xlim, ylim],
        cmap=pyplot.get_cmap("inferno"),
        norm=matplotlib.colors.LogNorm(),
        cmin=0.1
    )
    cbar = fig.colorbar(im, ax=ax)
    cbar.set_label("number of samples")

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    ax.set_xlabel("true redshift")
    ax.set_ylabel("best redshift estimate")

    ax.plot(
        [min(xlim[0],ylim[0]), max(xlim[1],ylim[1])],
        [min(xlim[0],ylim[0]), max(xlim[1],ylim[1])],
        'k:'
    )

    pyplot.tight_layout(h_pad=0, w_pad=0)
    pyplot.savefig(file_path)


def plot_outlierrate(file_path, weight, z_true, z_guess, risk, conf):
    weight = weight.astype(float)

    # judge whether outlier or not
    is_outlier = (numpy.abs((z_guess - z_true) / (z_true + 1.0)) > 0.15).astype(float)

    # sort by risk
    indices = numpy.argsort(risk)
    risk = risk[indices]
    is_outlier_risk = is_outlier[indices]
    weight_risk = weight[indices]

    # outlier rate by risk
    adoption_rate_risk = numpy.cumsum(weight_risk)
    outlier_rate_risk = numpy.cumsum(is_outlier_risk*weight_risk) / adoption_rate_risk

    # sort by conf
    indices = numpy.argsort(-conf)
    conf = conf[indices]
    is_outlier_conf = is_outlier[indices]
    weight_conf = weight[indices]

    # outlier rate by conf
    adoption_rate_conf = numpy.cumsum(weight_conf)
    outlier_rate_conf = numpy.cumsum(is_outlier_conf*weight_conf) / adoption_rate_conf

    # outlier rate in total
    outlier_rate_intotal = outlier_rate_conf[-1]
    variance = numpy.sum(weight_conf*numpy.square(is_outlier_conf - outlier_rate_intotal))
    sww = numpy.sum(numpy.square(weight_conf))
    variance_of_mean = sww / (1 - sww) * variance

    print "Outlier rate:", outlier_rate_intotal, "+-", numpy.sqrt(variance_of_mean)

    from matplotlib import pyplot
    import matplotlib.colors

    pyplot.rcParams.update({
        'axes.labelsize': 10,
        'font.size': 10,
        'legend.fontsize': 10,
        'xtick.labelsize': 10,
        'ytick.labelsize': 10,
        'font.family': "serif",
    })

    fig = pyplot.figure(figsize=(5.5, 5.5), dpi=600)

    #
    # outlier rate  to  adoption rate
    #
    ax = fig.add_subplot(2,2,1)

    ymax = outlier_rate_risk[-1]
    ymax = numpy.ceil(ymax * 100) / 100

    ax.set_xlim([0,100]); ax.set_xlabel("adoption rate (%)")
    ax.set_ylim([0,100*ymax]); ax.set_ylabel("outlier rate (%)")
    ax.plot(100*adoption_rate_conf, 100*outlier_rate_conf, "k--", label="select by conf.")
    ax.plot(100*adoption_rate_risk, 100*outlier_rate_risk, "k-" , label="select by risk")
    ax.legend(loc="upper left", frameon=False, framealpha=0)

    #
    # outlier rate  to  threshold
    #
    ax = fig.add_subplot(2,2,2)
    axt = ax.twiny()

    ax.set_ylim([0,100*ymax]); ax.set_ylabel("outlier rate (%)")

    axt.set_xlim([0,1]); axt.set_xlabel("risk threshold")
    ax .set_xlim([1,0]); ax .set_xlabel("conf. threshold")
    axt.plot(risk, 100*outlier_rate_risk, "k-" , label="risk")
    ax .plot(conf, 100*outlier_rate_conf, "k--", label="conf.")
    axt.legend(loc="upper left", frameon=False, framealpha=0)
    ax .legend(loc="lower right", frameon=False, framealpha=0)

    #
    # threshold  to  adoption rate
    #
    ax = fig.add_subplot(2,2,3)
    axt = ax.twinx()

    ax.set_xlim([0,100]); ax.set_xlabel("adoption rate (%)")

    axt.set_ylim([0,1]); axt.set_ylabel("risk threshold")
    ax .set_ylim([1,0]); ax .set_ylabel("conf. threshold")
    axt.plot(100*adoption_rate_risk, risk, "k-" , label="risk")
    ax .plot(100*adoption_rate_conf, conf, "k--", label="conf.")
    axt.legend(loc="lower right", frameon=False, framealpha=0)
    ax .legend(loc="upper left", frameon=False, framealpha=0)

    pyplot.tight_layout(h_pad=0, w_pad=0)
    pyplot.savefig(file_path)


def load_data(file_path, key):
    catalog = numpy.load(file_path)[key]

    ignored = set()
    idfield = "object_id"
    zfield = "redshift"
    wfield = "crossval_weights"

    ignored.update([idfield, zfield, wfield])

    xfields = []

    for key in catalog.dtype.descr:
        if (key[0] not in ignored
        and not key[0].endswith("_err")
        ):
            xfields.append(key)

    dimension = len(xfields)
    n = len(catalog)

    x = numpy.empty(shape=(n, dimension), dtype=float)
    dx = numpy.empty(shape=(n, dimension), dtype=float)
    z = catalog[zfield].astype(float)
    w = catalog[wfield].astype(float)
    object_id = catalog[idfield].astype(numpy.int64)

    ix = 0
    for key in xfields:
        x[:,ix] = catalog[key[0]]
        if key[0] + "_err" in catalog.dtype.names:
            dx[:,ix] = catalog[key[0]+"_err"]
        else:
            print "# Warning: input catalog does not include " + key[0] + "_err"
            dx[:,ix] = 0.0
        ix += 1

    isfinite = (
        numpy.all(numpy.isfinite(x), axis=1)
      & numpy.all(numpy.isfinite(dx), axis=1)
      & numpy.isfinite(z)
      & numpy.isfinite(w)
    )

    x = x[isfinite]
    dx = dx[isfinite]
    z = z[isfinite]
    w = w[isfinite]
    object_id = object_id[isfinite]

    w /= numpy.sum(w)

    return object_id, x, dx, z, w


if __name__ == "__main__":
    startup()
